use std::net::SocketAddr;
use std::sync::Arc;

use axum::{
    extract::{Query, State},
    http::{header::CONTENT_TYPE, StatusCode},
    response::{AppendHeaders, IntoResponse},
    routing::get,
    Json, Router,
};

use color_eyre::eyre::Result;

use clap::Parser;

use deadpool_postgres::{Config, ManagerConfig, Pool, RecyclingMethod, Runtime};
use tokio_postgres::NoTls;

use serde::Deserialize;

use tower_http::catch_panic::CatchPanicLayer;
use tower_http::trace::TraceLayer;

use tracing_subscriber::filter::{EnvFilter, LevelFilter};
use tracing_subscriber::fmt::format::FmtSpan;
use tracing_subscriber::prelude::*;

use tracing::{debug, info};

#[derive(Parser)]
struct CliConfig {
    #[clap(long, env)]
    bind_address: Option<SocketAddr>,

    #[clap(long, env, default_value = "postgres")]
    postgres_dbname: String,

    #[clap(long, env, default_value = "localhost")]
    postgres_host: String,

    #[clap(long, env)]
    postgres_password: String,

    #[clap(long, env, default_value = "5432")]
    postgres_port: u16,

    #[clap(long, env, default_value = "postgres")]
    postgres_user: String,

    #[clap(long, env, default_value = "mnemonics")]
    database_table: String,

    #[clap(long, env, default_value = "hash")]
    database_column: String,
}

struct AppState {
    pool: Pool,
    database_table: String,
    database_column: String,
}

fn setup_registry() {
    let envfilter = EnvFilter::builder()
        .with_default_directive(LevelFilter::DEBUG.into())
        .from_env_lossy();
    tracing_subscriber::registry()
        .with(envfilter)
        .with(tracing_subscriber::fmt::layer().with_span_events(FmtSpan::CLOSE))
        .with(tracing_error::ErrorLayer::default())
        .init();
}

async fn index() -> impl IntoResponse {
    (
        AppendHeaders([(CONTENT_TYPE, "text/html")]),
        include_str!("index.html"),
    )
}

#[tracing::instrument(skip(state))]
async fn check_hash(hash: &str, state: &AppState) -> Result<bool> {
    let client = state.pool.get().await?;
    let column = &state.database_column;
    let table = &state.database_table;
    let formatted_query = format!("SELECT {column} FROM {table} WHERE hash = $1");
    let query = client.prepare_cached(formatted_query.as_str()).await?;
    let rows = client.query(&query, &[&hash]).await?;
    if let Some(row) = rows.get(0) {
        let retrieved_hash: String = row.try_get(0)?;
        Ok(hash == retrieved_hash)
    } else {
        Ok(false)
    }
}

#[derive(Deserialize)]
struct CheckQuery {
    sha256: String,
}

// Note: Exposes *zero* information of potential errors to clients.
#[tracing::instrument(skip(query, state))]
async fn check_hash_slug(
    query: Query<CheckQuery>,
    State(state): State<Arc<AppState>>,
) -> (StatusCode, Json<Option<bool>>) {
    let result = check_hash(&query.sha256, &state).await;
    let status_code = match result.as_ref() {
        Ok(true) => StatusCode::OK,
        Ok(false) => StatusCode::NOT_FOUND,
        Err(e) => {
            debug!(%e, "Error while performing lookup");
            StatusCode::INTERNAL_SERVER_ERROR
        }
    };
    (status_code, Json(result.ok()))
}

#[tokio::main]
async fn main() -> Result<()> {
    setup_registry();
    color_eyre::install()?;

    let cli_config = CliConfig::parse();
    let mut config = Config::new();
    config.dbname = Some(cli_config.postgres_dbname);
    config.host = Some(cli_config.postgres_host);
    config.password = Some(cli_config.postgres_password);
    config.port = Some(cli_config.postgres_port);
    config.user = Some(cli_config.postgres_user);
    config.manager = Some(ManagerConfig {
        recycling_method: RecyclingMethod::Fast,
    });
    let pool = config.create_pool(Some(Runtime::Tokio1), NoTls)?;
    let addr = cli_config
        .bind_address
        .unwrap_or_else(|| (std::net::Ipv4Addr::new(127, 0, 0, 1), 8000).into());

    let app = Router::new()
        .route("/", get(index))
        .route("/check", get(check_hash_slug))
        .with_state(Arc::new(AppState {
            pool,
            database_table: cli_config.database_table,
            database_column: cli_config.database_column,
        }))
        .layer(CatchPanicLayer::new())
        .layer(TraceLayer::new_for_http());

    info!("server go nyoom");

    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await?;

    Ok(())
}
